from collections import OrderedDict
from unittest import TestCase
from flask import Flask, render_template, current_app
from flask_flacro import *


class FlacroTest(TestCase):
    def setUp(self):
        app = Flask(__name__)
        app.config['TESTING'] = True
        Flacro(app)
        self.app = app
        static_macro = FlacroFor(tag='static_macro', mwhere='test_macros/test.html', mname='test_macro_static')
        content_macro = FlacroFor(tag='content_macro', mwhere='test_macros/test.html',mname='test_macro_with_content')
        named_macro = FlacroFor(tag='named_macro', mwhere='test_macros/test.html',mname='test_named_macro')

        class TestSubclass(FlacroFor):
            def __init__(self, anything, **kwargs):
                self.anything=anything
                super(TestSubclass, self).__init__(tag=kwargs.get('tag', None),
                                                   mwhere='test_macros/test.html',
                                                   mname='test_macro_with_attr')

        attr_macro  = TestSubclass("abc", tag="ATTR_MACRO_1")
        attr_macro_alternate  = TestSubclass("123", tag="ATTR_MACRO_2")

        #accordian macro
        ai = []
        for i in range(5):
            ai.append(AccordianItem("accordian{}".format(i), FlacroFor(tag="inner_{}".format(i), mwhere='test_macros/test.html', mname='test_macro_tab_content', mattr={'iam': 'content'})))
        accordian_macro = AccordianGroupMacro("tag_for_accordian_macro", "accordian_test", ai)

        #tab group macro
        ti = OrderedDict()
        tie = TabItem('e', 'E', external="http://somewhere.com")
        tis = TabItem('s', 'S', static=FlacroFor(mwhere='test_macros/test.html', mname='test_macro_static'))
        tii = TabItem('i', 'I', independent=FlacroFor(mwhere='test_macros/test.html', mname='test_macro_independent', mattr={'iam': 'independent'}))
        tic = TabItem('c', 'C', content=FlacroFor(mwhere='test_macros/test.html', mname='test_macro_tab_content', mattr={'iam': 'content'}))
        tim = TabItem('m', 'M', minimal=FlacroFor(mwhere='test_macros/test.html', mname='test_macro_tab_content', mattr={'iam': 'minimal'}))
        tabitems = [tie, tis, tii, tic]
        [ti.update({k.label: k}) for k in tabitems]
        ti2 = ti
        ti2['m'] = tim
        tabs_macro = TabGroupMacro(minimal=False,
                                   tag="tag_for_tab_macro",
                                   tabs_label="test_tabs",
                                   tab_groups=ti)
        minimal_tabs_macro = TabGroupMacro(minimal=True,
                                           tag="tag_for_minimal_tabs",
                                           tabs_label="minimal_test_tabs",
                                           tab_groups=ti2)

        packaged_macros = {'accordian_macro': accordian_macro,
                           'tabs_macro': tabs_macro,
                           'minimal_tabs_macro': minimal_tabs_macro}

        @self.app.route('/static_macro')
        def stat_mac():
            return render_template('index.html', rstat=static_macro)

        @self.app.route('/content_macro')
        def con_mac():
            return render_template('index.html', rcont=content_macro, content={'any': "THING"})

        @self.app.route('/attr_macro')
        def sub_mac():
            return render_template('index.html', rattr=attr_macro)

        @self.app.route('/named_macro')
        def nam_mac():
            return render_template('index.html', test_named=True)

        @self.app.route('/<macro_type>')
        def gen_mac(macro_type):
            return render_template('index.html', m=packaged_macros[macro_type])

        @self.app.route('/one')
        def test_page_one():
            return "ONE"

        @self.app.route('/two')
        def test_page_two():
            return "TWO"

        #list/listitems macro
        l1 = LiItem('l1', kind='link', for_url='test_page_one')
        l2 = LiItem('l2', kind='link', for_url='test_page_two', arbitrary='for_attr_macro_route')
        l3 = LiItem('a plain item')

        links_list = ListMacro([l1, l2, l3])

        packaged_macros.update({'links_list': links_list})


    def tearDown(self):
        self.app = None



__version__ = '0.0.8'

__all__ = ('Flacro', 'FlacroFor', 'AccordianItem', 'AccordianGroupMacro',
    'TabItem', 'TabGroupMacro', 'ListMacro', 'LiItem')

from .flacro import (Flacro, FlacroFor)
from .packaged_macros import (AccordianItem, AccordianGroupMacro,
    TabItem, TabGroupMacro, ListMacro, LiItem)



from functools import partial
from .flacro import FlacroFor
from flask import url_for


class LiItem(FlacroFor):
    """A list item containing a link. Any kwargs left over will be passed to
    url creation

    :param li_tag:       a tag for item
    :param kind:         type of li item: plain(tag only), link, macro
    :param li:           is an actual html li, defaults True, set False to eliminate <li></li>
    :param css_class:    a css class
    :param for_url:      the flask route passed to url_for or a url
    :param external:     for_url is external
    """
    def __init__(self, li_tag, **kwargs):
        self.li_tag = li_tag
        self.kind = kwargs.pop('kind', 'plain')
        self.li = kwargs.pop('li', True)
        self.css_class = kwargs.pop('css_class', None)
        self._route = kwargs.pop('for_url', None)
        self._route_external = kwargs.pop('external', False)
        self._route_add = kwargs
        if self._route:
            self.url = partial(self.generate_url, self._route, self._route_add, external=self._route_external)
        super(LiItem, self).__init__(mname="{}item".format(self.kind), mwhere="macros/list.html")

    @staticmethod
    def generate_url(route, route_add, external=False):
        if external:
            return route
        else:
            return url_for(route, **route_add)


class ListMacro(FlacroFor):
    """A generalized list (links, macros, or anything)

    :param list_tag:    a tag for the lsit
    :param kind:        type of list ul or ol
    :param list_items:  a list of LiItems items
    :param css_class:   a css class
    :param css_id:      a css id
    """
    def __init__(self, list_items, **kwargs):
        self.list_tag = kwargs.get('list_tag', None)
        self.kind = kwargs.get('kind', 'ul')
        self.list_items = list_items
        self.css_class = kwargs.get('css_class', None)
        self.css_id = kwargs.get('css_id', None)
        super(ListMacro, self).__init__(mname='listmacro', mwhere="macros/list.html")


class AccordianItem(object):
    def __init__(self,
                 group_label,
                 interior,
                 is_open=True,
                 display_label=None):
        self.group_label = group_label
        self.interior = interior
        self.is_open = is_open
        self.of_accordian = None
        self.display_label = display_label

    @property
    def label(self):
        if self.display_label:
            return self.display_label
        else:
            return self.group_label


class AccordianGroupMacro(FlacroFor):
    def __init__(self,
                 tag,
                 accordian_label,
                 accordian_groups,
                 group_class="accordian-macro"):
        self.accordian_groups = accordian_groups
        self.accordian_label = accordian_label
        for g in accordian_groups:
            g.of_accordian = accordian_label
        self.accordian_groups = accordian_groups
        self.group_class = group_class
        super(AccordianGroupMacro, self).__init__(tag=tag,
                                                  mwhere="macros/accordian.html",
                                                  mname="accordian_group_macro")


class TabItem(object):
    def __init__(self, label, tab_label, **kwargs):
        self.label = label
        self.tab_label = tab_label
        self.set_tab(kwargs)

    def set_tab(self, kwargs):
        if kwargs:
            for k,v in kwargs.items():
                if k in ('minimal', 'external', 'static', 'independent', 'content'):
                    tab_type = k
                    tab_content = v
                else:
                    tab_type = 'content'
                    tab_content = "None"
                setattr(self, 'kind', tab_type)
                setattr(self, '_li', getattr(self, "make_li", None)(tab_type))
                setattr(self, tab_type, tab_content)

    def make_li(self, tab_type):
        tab = "{}_li".format(tab_type)
        return FlacroFor(mwhere="macros/tab.html",
                        mname=tab).renderable


class TabGroupMacro(FlacroFor):
    def __init__(self, **kwargs):
        self.tag = kwargs.get('tag', None)
        self.minimal = kwargs.get('minimal', False)
        if self.minimal:
            self.mname = "minimal_tabs"
        else:
            self.mname = "bootstrap_tabs"
        self.mwhere = kwargs.get('mwhere', "macros/tab.html")
        self.tabs_label = kwargs.get('tabs_label', None)
        self.tab_groups = kwargs.get('tab_groups', None)
        self.tabset_css_class = kwargs.get('tabset_css_class', "tabset-macro")
        self.vertical = kwargs.get('vertical', False)
        self.justified = kwargs.get('justified', False)
        self.pills =  kwargs.get('pills', False)
        super(TabGroupMacro, self).__init__(tag=self.tag,
                                            mname=self.mname,
                                            mwhere=self.mwhere,
                                            mattr={'tabs_label': self.tabs_label,
                                                   'tab_groups': self.tab_groups,
                                                   'tabs_css_class': self.tabset_css_class,
                                                   'tabs_nav_class': "tabbed-nav-{}".format(self.tabset_css_class),
                                                   'tabs_content_class': "content-for-tabs-{}".format(self.tabset_css_class),
                                                   'vertical': self.vertical,
                                                   'justified': self.justified,
                                                   'pills': self.pills,
                                                   })



import re
from flask import current_app, Blueprint, _app_ctx_stack
from werkzeug import LocalProxy, MultiDict, CombinedMultiDict
from .compat import with_metaclass
from collections import defaultdict
import weakref


_flacro_jinja = LocalProxy(lambda: current_app.jinja_env)
_glo = LocalProxy(lambda:  current_app.jinja_env.globals)

ATTR_BLACKLIST = re.compile("mwhere|mname|mattr|macros|^_")


class FlacroForMeta(type):
    def __new__(cls, name, bases, dct):
        new_class = super(FlacroForMeta, cls).__new__(cls, name, bases, dct)
        if not hasattr(cls, '_instances'):
            new_class._instances = defaultdict(weakref.WeakSet)
        if not hasattr(cls, '_manager'):
            cls._manager = {}
        cls._manager[new_class.__name__] = new_class
        return new_class

    def __init__(cls, name, bases, dct):
        if not hasattr(cls, '_registry'):
            cls._registry = {}
        else:
            cls._registry[name] = cls._instances
        super(FlacroForMeta, cls).__init__(name, bases, dct)


class FlacroFor(with_metaclass(FlacroForMeta)):
    """
    A container class for managing, holding and returning Jinja2 macros within
    a Flask application. Instance as-is or use as a mixin.

    m = FlacroFor(mwhere="macros/my_macro.html", mname="my_macro")

    class MyMacro(FlacroFor):
        def __init__(self, a, b):
            self.a = a
            self.b = b
            super(MyMacro, self).__init__(mwhere="macros/my_macro.html",
                                          mname="my_macro")

    where "macros/my_macro.html" is a file in your templates directory and
    "my_macro" is a defined macro within that file.

    :param mwhere:  the jinja template file location of your macro
    :param mname:   the name of the macro within the macro file
    :param mattr:   a dict of items you might want to access
                    e.g. {'a': 'AAAAAA', 'b': 'BBBBB'}
    :param macros:  a dict of macros within the same file specified
                    above as mwhere in the form {mname: mattr}
                    e.g. {'my_macro_1': {1: 'x', 2: 'y'},
                            'my_macro_2': None}
    """
    def __init__(self, **kwargs):
        self.tag = kwargs.get('tag', None)
        self.mwhere = kwargs.get('mwhere', None)
        self.mname = kwargs.get('mname', None)
        self._mattr = kwargs.get('mattr', None)
        self._macros = kwargs.get('macros', None)
        if self._mattr:
            for k, v in self._mattr.items():
                setattr(self, k, v)
        if self._macros:
            for k, v in self._macros.items():
                setattr(self, k, self.get_macro(k, mattr=v))
        self.register_instance(self)

    @classmethod
    def register_instance(cls, instance):
        if getattr(instance, 'tag', None):
            cls._instances[instance.tag] = weakref.ref(instance, instance) #viable or stupid
        else:
            cls._instances[None].add(instance)

    @property
    def ctx_prc(self):
        def ctx_prc(macro):
            return LocalProxy(lambda: getattr(macro, 'render', None))
        return {self.tag: ctx_prc(self)}

    def _public(self):
        return [k for k in self.__dict__.keys() if not ATTR_BLACKLIST.search(k)]

    @property
    def public(self):
        return {k: getattr(self, k, None) for k in self._public()}

    def update(self, **kwargs):
        [setattr(self, k, v) for k,v in kwargs.items()]

    def get_macro(self, mname, mattr=None, replicate=False):
        """returns another MacroFor instance with a differently named macro from
        the template location of this instance"""
        if replicate:
            mattr=self.public
        return FlacroFor(mwhere=self.mwhere,
                        mname=mname,
                        mattr=mattr)

    def jinja_template(self, mwhere):
        return _flacro_jinja.get_template(mwhere, globals=_glo).module

    def get_template_attribute(self, mwhere, mname):
        return getattr(self.jinja_template(mwhere), mname)

    @property
    def renderable(self):
        """the macro held but not called"""
        try:
            return self.get_template_attribute(self.mwhere, self.mname)
        except RuntimeError:
            return LocalProxy(lambda: self.get_template_attribute(self.mwhere, self.mname))

    @property
    def render(self):
        """calls the macro, passing itself as accessible within"""
        return self.renderable(self)

    @property
    def render_static(self):
        """calls the macro passing in no variable"""
        return self.renderable()

    def render_with(self, content):
        """calls the macro with the content specified as parameter(s)"""
        return self.renderable(content)

    def __repr__(self):
        return "<MacroFor {} ({}: {})>".format(getattr(self, 'tag', None), self.mwhere, self.mname)


class Flacro(object):
    """flask/jinja2 tools for managing template macros"""
    def __init__(self,
                 app=None,
                 register_blueprint=True):
        self.app = app
        self.register_blueprint = register_blueprint
        self._registry = FlacroFor._registry
        self._managed = FlacroFor._manager
        if self.app is not None:
            self.init_app(self.app)

    @property
    def provides(self):
        return CombinedMultiDict([(MultiDict([(k,v),(k, self._registry.get(k, None))]))
            for k,v in self._managed.items()])

    def init_app(self, app):
        app.extensions['flacro'] = self
        app.before_request(self.make_ctx_prc)
        if self.register_blueprint:
            app.register_blueprint(self._blueprint)

    def make_ctx_prc(self):
        [[self.app.jinja_env.globals.update(macro().ctx_prc)
            for m, macro in mf.items() if m]
            for mf in self._registry.values()]

    @property
    def _blueprint(self):
        return Blueprint('flacro', __name__, template_folder='templates')



"""
Flask-Macro4
-------------
A tool for working with template macros in flask/jinja2 plus assorted useful
pre-built macros
"""
from setuptools import setup

setup(
    name='Flask-Flacro',
    version='0.0.8',
    url='https://github.com/thrisp/flacro',
    license='MIT',
    author='hurrata/thrisp',
    author_email='blueblank@gmail.com',
    description='flask/jinja2 templating tools',
    long_description=__doc__,
    packages=['flask_flacro'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask >= 0.9',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    test_suite='nose.collector',
    tests_require=[
        'nose',
        'blinker'],
)



from __future__ import with_statement

import sys
import os
from flask import Flask, render_template, current_app

from tests import *


class FlacroBaseCase(FlacroTest):
    def test_bases(self): pass

    def test_macro_static(self):
        rv = self.app.test_client().get('/static_macro')
        self.assertIn(b'STATIC MACRO', rv.data)

    def test_macro_content(self):
        rv = self.app.test_client().get('/content_macro')
        self.assertIn(b'THING', rv.data)

    def test_macro_attr(self):
        rv = self.app.test_client().get('/attr_macro')
        self.assertIn(b'abc', rv.data)

    def test_macro_named(self):
        rv = self.app.test_client().get('/named_macro')
        self.assertIn(b'A NAMED MACRO RENDERED BY NAME', rv.data)
        self.assertIn(b'123', rv.data)


class PackagedMacrosTestCase(FlacroTest):
    def test_list(self):
        rv = self.app.test_client().get('/links_list')
        self.assertIn(b'l1', rv.data)
        self.assertIn(b'href="/one"', rv.data)
        self.assertIn(b'arbitrary=for_attr_macro_route', rv.data)
        self.assertIn(b'a plain item', rv.data)

    def test_accordian(self):
        rv = self.app.test_client().get('/accordian_macro')
        self.assertIn(b'accordian_test-accordion', rv.data)
        self.assertIn(b'#accordian_test', rv.data)

    def test_tabs(self):
        rv = self.app.test_client().get('/tabs_macro')
        self.assertIn(b'tabset-macro', rv.data)
        self.assertIn(b'tab-content', rv.data)
        rv = self.app.test_client().get('minimal_tabs_macro')
        self.assertIn(b'tabset', rv.data)
        self.assertIn(b'minimal', rv.data)


if __name__ == '__main__':
    unittest.main()
